/**
 * @file vaapi_x11.c
 * @brief VA API using X11 video output module for VLC media player
 */
/*****************************************************************************
 * Copyright © 2011, M2X BV
 * Author: Jean-Paul Saman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <assert.h>

#include <vlc_common.h>
#include <vlc_plugin.h>
#include <vlc_picture.h>
#include <vlc_vout_display.h>
#include <vlc_vout_window.h>
#include <vlc_picture_pool.h>

#ifdef HAVE_AVCODEC_VAAPI

#ifdef HAVE_LIBAVCODEC_AVCODEC_H
#   include <libavcodec/avcodec.h>
#else
#   include <avcodec.h>
#endif

#include "avcodec.h"

#include <vlc_xlib.h>
#include <libavcodec/vaapi.h>
#include <va/va.h>
#include <va/va_x11.h>
#include "va.h"
#include "vaapi.h"

#include <X11/Xutil.h>

struct vout_display_sys_t
{
    vlc_va_conn_t *conn;

    vout_window_t *embed;

    picture_pool_t *pool; /* picture pool */

    unsigned int i_display_order;
    bool visible; /* whether to draw */
};

static picture_pool_t *Pool (vout_display_t *, unsigned);
static void DisplayPicture (vout_display_t *, picture_t *, subpicture_t *);
static int Control  (vout_display_t *, int, va_list);

/* */
static int FindVAFourCC(vout_display_t *vd)
{
    vout_display_sys_t *sys = (vout_display_sys_t *) vd->sys;

    /* Find and create a supported image chroma */
    int i_fmt_count = vaMaxNumImageFormats(sys->conn->p_display);
    VAImageFormat *p_fmt = calloc(i_fmt_count, sizeof(*p_fmt));
    if( !p_fmt )
        goto error;

    if( vaQueryImageFormats( sys->conn->p_display, p_fmt, &i_fmt_count ) )
    {
        free( p_fmt );
        goto error;
    }

    vlc_fourcc_t  i_chroma = 0;
    int32_t i_bits_per_pixel = 0;
    for( int i = 0; i < i_fmt_count; i++ )
    {
        if( p_fmt[i].fourcc == VA_FOURCC( 'Y', 'V', '1', '2' ) ||
            p_fmt[i].fourcc == VA_FOURCC( 'I', '4', '2', '0' ) ||
            p_fmt[i].fourcc == VA_FOURCC( 'N', 'V', '1', '2' ) )
        {
            i_chroma = VLC_CODEC_YV12;
            i_bits_per_pixel = p_fmt[i].bits_per_pixel;
            break;
        }
    }
    free( p_fmt );
    if( !i_chroma )
        goto error;

    vd->fmt.i_chroma = i_chroma;
    vd->fmt.i_bits_per_pixel = i_bits_per_pixel;
    return VLC_SUCCESS;

error:
    return VLC_EGENERIC;
}

static vout_window_t *MakeWindow (vout_display_t *vd)
{
    vout_window_cfg_t wnd_cfg;

    memset (&wnd_cfg, 0, sizeof (wnd_cfg));
    wnd_cfg.type = VOUT_WINDOW_TYPE_XID;
    wnd_cfg.x = var_InheritInteger (vd, "video-x");
    wnd_cfg.y = var_InheritInteger (vd, "video-y");
    wnd_cfg.width  = vd->cfg->display.width;
    wnd_cfg.height = vd->cfg->display.height;

    vout_window_t *wnd = vout_display_NewWindow (vd, &wnd_cfg);
    if (wnd == NULL)
        msg_Err (vd, "parent window not available");
    return wnd;
}

/**
 * Probe the X server.
 */
int OpenVaapiX11 (vlc_object_t *obj)
{
    vout_display_t *vd = (vout_display_t *) obj;

    if( !vlc_xlib_init( obj ) )
        return VLC_EGENERIC;

    vout_display_sys_t *sys = malloc(sizeof(*sys));
    if (unlikely(sys == NULL))
        return VLC_ENOMEM;

    vd->sys = sys;

    sys->conn = NULL;
    sys->pool = NULL;
    sys->visible = false;
    sys->i_display_order = 0;

    sys->embed = MakeWindow(vd);
    if (unlikely(sys->embed == NULL))
        goto error;

    /* Create a VA display */
    sys->conn = vlc_va_Initialize(sys->embed->display.x11);
    if (!sys->conn)
        goto error;

    if (FindVAFourCC(vd) != VLC_SUCCESS)
        goto error;

    msg_Dbg(vd, "using VAAPI X11 window (version %d.%d)",
            sys->conn->i_version_major, sys->conn->i_version_minor);

    /* */
    vout_display_info_t info = vd->info;
    info.is_slow = false;
    info.has_pictures_invalid = false;
    info.has_event_thread = false;
    info.has_hide_mouse = false;
    info.subpicture_chromas = &vd->fmt.i_chroma;

    /* Setup vout_display_t once everything is fine */
    vd->info = info;

    vd->pool = Pool;
    vd->prepare = NULL; //PictureRender;
    vd->display = DisplayPicture;
    vd->control = Control;
    vd->manage = NULL; //Manage;

    /* */
    bool is_fullscreen = vd->cfg->is_fullscreen;
    if (is_fullscreen && vout_window_SetFullScreen (sys->embed, true))
        is_fullscreen = false;
    vout_display_SendEventFullscreen (vd, is_fullscreen);
    vout_display_SendEventDisplaySize (vd, vd->cfg->display.width, vd->cfg->display.height, is_fullscreen);

    return VLC_SUCCESS;

error:
    CloseVaapiX11(obj);
    return VLC_EGENERIC;
}

void CloseVaapiX11 (vlc_object_t *obj)
{
    vout_display_t *vd = (vout_display_t *) obj;
    vout_display_sys_t *sys = (vout_display_sys_t *) vd->sys;

    if (sys->conn)
    {
        if (sys->conn->p_display_x11)
        {
            XFlush(sys->conn->p_display_x11);
            XSync(sys->conn->p_display_x11, False);
        }
    }
    if (sys->embed)
        vout_display_DeleteWindow (vd, sys->embed);
    if (sys->conn)
        vlc_va_Terminate(sys->conn);

    if (sys->pool)
    {
        picture_pool_Delete(sys->pool);
        sys->pool = NULL;
    }
    free(vd->sys);
}

static int Control (vout_display_t *vd, int query, va_list ap)
{
    vout_display_sys_t *sys = vd->sys;

    switch (query)
    {
    case VOUT_DISPLAY_CHANGE_FULLSCREEN:
    {
        const vout_display_cfg_t *c = va_arg (ap, const vout_display_cfg_t *);
        return vout_window_SetFullScreen (sys->embed, c->is_fullscreen);
    }

    case VOUT_DISPLAY_CHANGE_WINDOW_STATE:
    {
        unsigned state = va_arg (ap, unsigned);
        return vout_window_SetState (sys->embed, state);
    }

    case VOUT_DISPLAY_CHANGE_DISPLAY_SIZE:
    case VOUT_DISPLAY_CHANGE_ZOOM:
    case VOUT_DISPLAY_CHANGE_DISPLAY_FILLED:
    case VOUT_DISPLAY_CHANGE_SOURCE_ASPECT:
    case VOUT_DISPLAY_CHANGE_SOURCE_CROP:
    {
        /* I am not sure it is always necessary, but it is way simpler ... */
        /* FIXME: adjust picture size */
        const vout_display_cfg_t *cfg;
        const video_format_t *source;
        bool is_forced = false;

        if (query == VOUT_DISPLAY_CHANGE_SOURCE_ASPECT
         || query == VOUT_DISPLAY_CHANGE_SOURCE_CROP)
        {
            source = (const video_format_t *)va_arg (ap, const video_format_t *);
            cfg = vd->cfg;
        }
        else
        {
            source = &vd->source;
            cfg = (const vout_display_cfg_t*)va_arg (ap, const vout_display_cfg_t *);
            if (query == VOUT_DISPLAY_CHANGE_DISPLAY_SIZE)
                is_forced = (bool)va_arg (ap, int);
        }

        /* */
        if (query == VOUT_DISPLAY_CHANGE_DISPLAY_SIZE
         && is_forced
         && (cfg->display.width  != vd->cfg->display.width
           ||cfg->display.height != vd->cfg->display.height)
         && vout_window_SetSize (sys->embed,
                                 cfg->display.width, cfg->display.height))
            return VLC_EGENERIC;

        return VLC_SUCCESS;
    }
    case VOUT_DISPLAY_RESET_PICTURES:
    {
        assert(0);
        return VLC_SUCCESS;
    }

    /* Hide the mouse. It will be send when
     * vout_display_t::info.b_hide_mouse is false */
    case VOUT_DISPLAY_HIDE_MOUSE:
        /* FIXME */
        return VLC_SUCCESS;

    default:
        msg_Err (vd, "Unknown request in XCB vout display");
        return VLC_EGENERIC;
    }
}

static void PictureRelease(picture_t *pic)
{
    assert(pic);
    assert(sizeof(pic->p_sys) == sizeof(picture_sys_t));

    if (pic->p_sys->surface)
    {
        vlc_mutex_lock(&pic->p_sys->surface->lock);
        pic->p_sys->surface->i_refcount--;
        assert(pic->p_sys->surface->i_refcount >= 0);
        vlc_mutex_unlock(&pic->p_sys->surface->lock);
        pic->p_sys->surface = NULL; /* FIXME: needed ? */
    }
}

static void DisplayPicture(vout_display_t *vd, picture_t *pic, subpicture_t *subpicture)
{
    vout_display_sys_t *sys = vd->sys;
    VLC_UNUSED(subpicture);

#if 0
    if (subpicture)
    {
        /* FIXME: blend subpicture into VAAPI structure */
    }
#endif

    assert(pic->p_sys);
    if (pic->p_sys->surface == NULL)
    {
        msg_Err(vd, "Discarding picture - no surface information");
        picture_Release(pic);
        return;
    }

    unsigned int i_order = (sys->i_display_order > 0) ?
        sys->i_display_order++ : pic->p_sys->surface->i_order;
    if (i_order != pic->p_sys->surface->i_order)
    {
        /* Reset picture */
        vlc_mutex_lock(&pic->p_sys->surface->lock);
        msg_Err(vd, "Reclaimed picture - id=%d, order=%d, refcount=%d (%p)",
                pic->p_sys->surface->i_id, pic->p_sys->surface->i_order,
                pic->p_sys->surface->i_refcount, pic);
        if (pic->p_sys->surface->i_refcount > 1)
            pic->p_sys->surface->i_refcount = 1;
        vlc_mutex_unlock(&pic->p_sys->surface->lock);
        pic->p_sys->surface = NULL;
        picture_Release(pic);
        return;
    }

    vout_display_place_t place;
    vout_display_PlacePicture (&place, &vd->source, vd->cfg, false);

    VASurfaceID i_surface_id = pic->p_sys->surface->i_id;
    VAStatus status = vaPutSurface(sys->conn->p_display,
                                   i_surface_id,
                                   sys->embed->handle.xid,
                                   pic->format.i_y_offset, pic->format.i_x_offset,
                                   pic->format.i_visible_width, pic->format.i_visible_height,
                                   place.y, place.x, place.width, place.height,
                                   NULL, 0, /* clipping */
                                   /* NOTE: vaCreateContext() specifies VA_PROGRESSIVE */
#if 0
                                   VA_FRAME_PICTURE);
#else
                                   pic->b_progressive ? VA_FRAME_PICTURE :
                                       (pic->b_top_field_first ? VA_TOP_FIELD : VA_BOTTOM_FIELD));
#endif
    if (status != VA_STATUS_SUCCESS)
        msg_Err(vd, "failed displaying picture (id=%d) %d", i_surface_id, status);
#if 0
    else
        msg_Info(vd, "displaying picture (id=%d, order=%d, refcount=%d)",
                 pic->p_sys->surface->i_id, pic->p_sys->surface->i_order,
                 pic->p_sys->surface->i_refcount);
#endif
    if ((status != VA_STATUS_ERROR_INVALID_SURFACE) &&
        (status != VA_STATUS_ERROR_INVALID_DISPLAY))
        PictureRelease(pic);
    picture_Release(pic);
}

/**
 * Return a direct buffer
 */
static picture_pool_t *Pool (vout_display_t *vd, unsigned requested_count)
{
    vout_display_sys_t *sys = vd->sys;

    assert(requested_count > 0);
    if (sys->pool)
        return sys->pool;

    picture_t **pic_array = (picture_t **) calloc(requested_count, sizeof(picture_t *));
    if (!pic_array)
        return NULL;

    unsigned int count;
    for (count = 0; count < requested_count; count++)
    {
        picture_t *pic = (picture_t*) picture_NewFromFormat(&vd->fmt);
        if (!pic)
            break;
        assert(pic->p_sys == NULL);
        picture_sys_t *sys = (picture_sys_t *) calloc(1, sizeof(picture_sys_t));
        if (!sys)
        {
            picture_Release(pic);
            break;
        }
        pic_array[count] = pic;
        pic_array[count]->b_progressive = true;
        pic_array[count]->p_sys = sys;
    }

    if (count == 0)
        goto error;

    sys->pool = picture_pool_New(count, pic_array);
    if (!sys->pool)
        goto error;

    free(pic_array);
    return sys->pool;

error:
    /* Cleanup on error */
    for (unsigned int i = 0; i < count; i++)
    {
        picture_t *pic = pic_array[i];
        if (pic)
            picture_Release(pic);
    }
    free(pic_array);
    return NULL;
}

#endif /* HAVE_AVCODEC_VAAPI */
